---json
{
  "title": "Some things I learned from building my first WebComponents",
  "date": "2024-12-13T11:20:04.269Z",
  "tags": [
    "webcomponents",
    "javascript",
    "html",
    "draft"
  ],
  "excerpt": "My team at work have recently released the first versions of some new WebComponents for internal use. As I'd only had fairly limited experience with WebComponents before starting this, I learned a few things that I thought I'd share."
}
---

My team at work have recently released the first versions of some new WebComponents for internal use. As I'd only had fairly limited experience with WebComponents before starting this, I learned a few things that I thought I'd share.

Our first component was an implementation of the ComboBox pattern - basically a custom Select element, with some brand-specific styling and functionality. We were doing this in an effort to remove the many custom versions scattered around our product, especially those that hadn't been built with accessibility in mind. We decided to use [Lit](https://lit.dev) to build the components, as it's fairly lightweight but handles a lot of the boilerplate while still exposing the native APIs.

## It's not React, so don't try to emulate it

Pretty much all of the frontend code currently uses React, which means that it's not unusual to see components with these sorts of APIs:

```javascript
<SomeComponent
  title="Something"
  icon={<IconComponent />}
  someFlag={true}
  complexPropertyThatIsUsedToRender={[
    { foo: 'foo', bar: 'bar' },
    ...
  ]}
/>
```

So initially, we tried to emulate this behaviour for our ComboBox component, which you _can_ do with Lit by using the `json` type for taking complex property types. I have no idea if it's possible to pass a DOM element as a property, and frankly I don't want to find out. But this is an anti-pattern, because we weren't building a React component, we're defining custom HTML elements - and they should be consumed just as you would an HTML element.

So instead of doing that, we used [slots](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/slot) for user-defined content. In our case, that included the `option` elements, which we applied to a named slot:

```html
<our-custom-combobox id="some-id">
  <span slot="label">The label text</span>
  
  <option slot="option" value="value">Option 1</option>
  <option slot="option" value="value2">Option 2</option>
  ...
</our-custom-combobox>
```

Then, as I mentioned in my [previous post about this pattern](/post/til-using-lit-decorators-for-better-webcomponent-apis/), we use the `queryAssignedElements` Lit decorator to retrieve and "enhance" the options for our use-case. This meant our component started to feel a bit more like a native element, which is what we ultimately wanted to go for.

## ElementInternals lets the component act like a regular form input

We found some cases where the label for our component needed to be external to it, for structural or formatting reason. Of course, this doesn't work for WebComponents because IDs are scoped the shadow root, so the browser isn't aware that the id on the label `for` attribute refers to the element in the WebComponent. 

So, after putting out [a question on Mastodon](https://social.lol/@lewis/113599779828387002) about this, thankfully [Aleš Roubíček had the answer we needed](https://indieweb.social/@alesroubicek/113599829106390768). The [ElementInternals api](https://developer.mozilla.org/en-US/docs/Web/API/ElementInternals) lets you call `this.attachInternals()`, and allows you to associate the component with a form, as well as other elements (including labels!). I had to make sure `delegatesFocus` was true when calling `attachShadow`, but once I did that my component was correctly associated with the form and labels.

I just had to make sure that my component was calling `this.internals.setFormValue()` whenever the value was changed, and then it would also send the value and name as part of the `FormData` when submitting. It can even take part in native form validation via the `ElementInternals.validity` object. The only thing that it can't do is match the `:user-invalid` selector, because that hasn't yet been implemented for any browser.

## Shadow Parts make stylable WebComponents much easier
